import numpy as np


def secondorder47_1(nc, group_indexes, subgroups, vki, eps_kl, secondorder):
    # condition between group 'COO-' and 'CH3'
    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond1 = 'CH3' in subgroups_id_molecule and 'COO-' in subgroups_id_molecule
        cond10 = subgroups_id_molecule.shape[0] == 2
        if cond1 and cond10:

            cond1_in1 = np.where(subgroups_id_molecule == 'CH3')
            bool11 = vki_molecule[cond1_in1] == 1

            cond1_in2 = np.where(subgroups_id_molecule == 'COO-')
            bool12 = vki_molecule[cond1_in2] == 1

            if bool11 and bool12:

                index1 = cond1_in1[0] + index0

                index47 = np.where(subgroups == 47)[0]

                boolk = secondorder.group_k == 'CH3'
                booll = secondorder.group_l == 'COO-'
                eps147 = secondorder[boolk & booll]['eps_kl'].values

                eps_kl[index1, index47] = eps147
                eps_kl[index47, index1] = eps147

            del bool11, bool12


def secondorder47_2(nc, group_indexes, subgroups, vki, eps_kl, secondorder):
    # condition between group 'COO-' and 'CH2'
    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond2 = 'CH3' in subgroups_id_molecule and 'COO-' in subgroups_id_molecule and 'CH2' in subgroups_id_molecule
        cond20 = subgroups_id_molecule.shape[0] == 3
        if cond2 and cond20:

            cond2_in1 = np.where(subgroups_id_molecule == 'CH3')
            bool21 = vki_molecule[cond2_in1] == 1

            cond2_in2 = np.where(subgroups_id_molecule == 'CH2')
            bool22 = vki_molecule[cond2_in2] == 1

            cond2_in3 = np.where(subgroups_id_molecule == 'COO-')
            bool23 = vki_molecule[cond2_in3] == 1

            if bool21 and bool22 and bool23:

                index2 = cond2_in2[0] + index0

                index47 = np.where(subgroups == 'COO-')[0]
                boolk = secondorder.group_k == 'CH2'
                booll = secondorder.group_l == 'COO-'
                eps247 = secondorder[boolk & booll]['eps_kl'].values

                eps_kl[index2, index47] = eps247
                eps_kl[index47, index2] = eps247

            del bool21, bool22, bool23


def secondorder19_14(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                     secondorder, subgroup_id_asso, molecule_id_index_asso,
                     sites_cumsum, epsAB_kl, kAB_kl, secondasso):
    # condition between group 'CH2OH' and 'H2O'
    bool19 = []

    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond3 = 'CH2OH' in subgroups_id_molecule

        cond30 = subgroups_id_molecule.shape[0] == 3

        cond31 = subgroups_id_molecule.shape[0] == 2

        cond32 = 'CH2' in subgroups_id_molecule

        if cond3 and cond30 and cond32:

            cond3_in19 = np.where(subgroups_id_molecule == 'CH2OH')
            bool319 = vki_molecule[cond3_in19] == 1

            cond3_in2 = np.where(subgroups_id_molecule == 'CH2')
            bool32 = vki_molecule[cond3_in2] in [1, 2]

            whereR = np.logical_and(subgroups_id_molecule != 'CH2OH', subgroups_id_molecule != 'CH2')
            R_id = subgroups_id_molecule[whereR]

            if R_id in ['CH3', 'NH2', 'NH', 'N']:
                bool33 = vki_molecule[whereR] == 1
            else:
                bool33 = False

            if R_id == 'CH3' and vki_molecule[cond3_in2] == 2:
                # for butanol
                bool34 = False
            else:
                bool34 = True

        elif cond3 and cond31:

            cond3_in19 = np.where(subgroups_id_molecule == 'CH2OH')
            bool319 = vki_molecule[cond3_in19] == 1

            bool32 = True
            whereR = np.where(subgroups_id_molecule != 'CH2OH')
            R_id = subgroups_id_molecule[whereR]
            bool34 = True

            if R_id in ['CH3', 'NH2', 'NH', 'N']:
                bool33 = vki_molecule[whereR] == 1
            else:
                bool33 = False

        else:
            bool319 = False
            bool32 = False
            bool33 = False
            bool34 = False

        bool19.append(bool319 and bool32 and bool33 and bool34)

        if bool319 and bool32 and bool33 and bool34:

            index19 = cond3_in19[0] + index0

            index14 = np.where(subgroups == 'H2O')[0]

            boolk = secondorder.group_k == 'CH2OH'
            booll = secondorder.group_l == 'H2O'
            eps1914 = secondorder[boolk & booll]['eps_kl'].values

            eps_kl[index19, index14] = eps1914
            eps_kl[index14, index19] = eps1914

    bool19 = np.hstack(bool19)
    cond_asso19_0 = np.any(bool19)
    cond_asso19_1 = np.any(subgroup_id_asso == 'CH2OH')

    if cond_asso19_0 and cond_asso19_1:
        where14 = np.where(subgroup_id_asso == 'H2O')

        where19all = np.where(subgroup_id_asso == 'CH2OH')
        where19all_id = molecule_id_index_asso[where19all]
        where19second = np.array(bool19)[where19all_id]
        where19 = where19all[0][where19second]
        ncond19 = len(where19)

        boolk = secondasso.group_k == 'CH2OH'
        booll = secondasso.group_l == 'H2O'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]

        for k in range(ncond19):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values

                if siteK == 'H' and siteL == 'e1':
                    indexH = sites_cumsum[where14] + 1
                    indexOH = sites_cumsum[where19[k]]

                    epsAB_kl[indexH, indexOH] = epsAB
                    epsAB_kl[indexOH, indexH] = epsAB
                    kAB_kl[indexH, indexOH] = kAB
                    kAB_kl[indexOH, indexH] = kAB
                elif siteK == 'e1' and siteL == 'H':

                    indexH = sites_cumsum[where14]
                    indexOH = sites_cumsum[where19[k]] + 1

                    epsAB_kl[indexH, indexOH] = epsAB
                    epsAB_kl[indexOH, indexH] = epsAB
                    kAB_kl[indexH, indexOH] = kAB
                    kAB_kl[indexOH, indexH] = kAB


def secondorder21(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                  secondorder, subgroup_id_asso, molecule_id_index_asso,
                  sites_cumsum, epsAB_kl, kAB_kl, secondasso):
    # condition between groups 'NH2'-'H2O' and 'NH2'-'CO2'

    bool21 = []
    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond4 = 'NH2' in subgroups_id_molecule and 'CH2' in subgroups_id_molecule and 'CH2OH' in subgroups_id_molecule
        cond40 = subgroups_id_molecule.shape[0] == 3

        bool419, bool421, bool42 = False, False, False

        if cond4 and cond40:

            cond4_in19 = np.where(subgroups_id_molecule == 'CH2OH')
            bool419 = vki_molecule[cond4_in19] == 1

            cond4_in21 = np.where(subgroups_id_molecule == 'NH2')
            bool421 = vki_molecule[cond4_in21] == 1

            cond4_in2 = np.where(subgroups_id_molecule == 'CH2')
            bool42 = vki_molecule[cond4_in2] in [1, 2]

        bool21.append(bool419 and bool421 and bool42)
        if bool419 and bool421 and bool42:

            index21 = cond4_in21[0] + index0

            index14 = np.where(subgroups == 'H2O')[0]

            index17 = np.where(subgroups == 'CO2')[0]

            boolNH2 = secondorder.group_k == 'NH2'
            boolH2O = secondorder.group_l == 'H2O'
            boolCO2 = secondorder.group_l == 'CO2'
            eps2114 = secondorder[boolNH2 & boolH2O]['eps_kl'].values
            eps2117 = secondorder[boolNH2 & boolCO2]['eps_kl'].values
            lr2117 = secondorder[boolNH2 & boolCO2]['lr_kl'].values

            eps_kl[index21, index14] = eps2114
            eps_kl[index14, index21] = eps2114

            eps_kl[index17, index21] = eps2117
            eps_kl[index21, index17] = eps2117

            lr_kl[index17, index21] = lr2117
            lr_kl[index21, index17] = lr2117

        del bool419, bool421, bool42

    bool21 = np.hstack(bool21)
    cond_asso21_0 = np.any(bool21)
    cond_asso21_14 = np.any(subgroup_id_asso == 'H2O')
    cond_asso21_17 = np.any(subgroup_id_asso == 'CO2')
    cond_asso21_19 = np.any(subgroup_id_asso == 'CH2OH')

    where21all = np.where(subgroup_id_asso == 'NH2')
    where21all_id = molecule_id_index_asso[where21all]
    where21second = bool21[where21all_id]
    where21 = where21all[0][where21second]
    ncond21 = len(where21)

    if cond_asso21_0 and cond_asso21_14:

        where14 = np.where(subgroup_id_asso == 'H2O')

        boolk = secondasso.group_k == 'NH2'
        booll = secondasso.group_l == 'H2O'
        df = secondasso[boolk & booll]

        len1 = df.shape[0]

        for k in range(ncond21):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values
                if siteK == 'H' and siteL == 'e1':
                    indexH = sites_cumsum[where14] + 1
                    indexNH2 = sites_cumsum[where21[k]]

                    epsAB_kl[indexH, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexH] = epsAB
                    kAB_kl[indexH, indexNH2] = kAB
                    kAB_kl[indexNH2, indexH] = kAB

                elif siteK == 'e1' and siteL == 'H':

                    indexH = sites_cumsum[where14]
                    indexNH2 = sites_cumsum[where21[k]] + 1

                    epsAB_kl[indexH, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexH] = epsAB
                    kAB_kl[indexH, indexNH2] = kAB
                    kAB_kl[indexNH2, indexH] = kAB


    if cond_asso21_0 and cond_asso21_17:

        where17 = np.where(subgroup_id_asso == 'CO2')

        boolk = secondasso.group_k == 'NH2'
        booll = secondasso.group_l == 'CO2'
        df = secondasso[boolk & booll]

        len1 = df.shape[0]

        for k in range(ncond21):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values
                if siteK == 'e1' and siteL == 'e1':
                    indexCO2 = sites_cumsum[where17]
                    indexNH2 = sites_cumsum[where21[k]] + 1

                    epsAB_kl[indexCO2, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexNH2] = kAB
                    kAB_kl[indexNH2, indexCO2] = kAB
                elif siteK == 'e1' and siteL == 'e2':
                    indexCO2 = sites_cumsum[where17] + 1
                    indexNH2 = sites_cumsum[where21[k]] + 1

                    epsAB_kl[indexCO2, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexNH2] = kAB
                    kAB_kl[indexNH2, indexCO2] = kAB

    if cond_asso21_0 and cond_asso21_19:

        where19 = np.where(subgroup_id_asso == 'CH2OH')

        boolk = secondasso.group_k == 'NH2'
        booll = secondasso.group_l == 'CH2OH'
        df = secondasso[boolk & booll]

        len1 = df.shape[0]

        for k in range(ncond21):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values

                if siteK == 'H' and siteL == 'e1':
                    indexOH = sites_cumsum[where19] + 1
                    indexNH2 = sites_cumsum[where21[k]]

                    epsAB_kl[indexOH, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexOH] = epsAB
                    kAB_kl[indexOH, indexNH2] = kAB
                    kAB_kl[indexNH2, indexOH] = kAB
                elif siteK == 'e1' and siteL == 'H':

                    indexOH = sites_cumsum[where19]
                    indexNH2 = sites_cumsum[where21[k]] + 1

                    epsAB_kl[indexOH, indexNH2] = epsAB
                    epsAB_kl[indexNH2, indexOH] = epsAB
                    kAB_kl[indexOH, indexNH2] = kAB
                    kAB_kl[indexNH2, indexOH] = kAB


def secondorder22(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                  secondorder, subgroup_id_asso, molecule_id_index_asso,
                  sites_cumsum, epsAB_kl, kAB_kl, secondasso):
    # condition between groups 'NH'-'H2O' and 'NH'-'CO2'
    bool22 = []
    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond6 = 'NH' in subgroups_id_molecule and 'CH2' in subgroups_id_molecule and 'CH2OH' in subgroups_id_molecule
        cond60 = subgroups_id_molecule.shape[0] == 3

        bool622, bool62, bool619 = False, False, False

        if cond6 and cond60:

            cond6_in22 = np.where(subgroups_id_molecule == 'NH')
            bool622 = vki_molecule[cond6_in22] == 1

            cond6_in2 = np.where(subgroups_id_molecule == 'CH2')
            bool62 = vki_molecule[cond6_in2] == 2

            cond6_in19 = np.where(subgroups_id_molecule == 'CH2OH')
            bool619 = vki_molecule[cond6_in19] == 2

        bool22.append(bool622 and bool62 and bool619)
        if bool622 and bool62 and bool619:
            # second order mie between NH and CO2
            index22 = cond6_in22[0] + index0

            index17 = np.where(subgroups == 'CO2')[0]

            boolNH = secondorder.group_k == 'NH'
            boolCO2 = secondorder.group_l == 'CO2'
            eps2217 = secondorder[boolNH & boolCO2]['eps_kl'].values
            lr2217 = secondorder[boolNH & boolCO2]['lr_kl'].values

            eps_kl[index17, index22] = eps2217
            eps_kl[index22, index17] = eps2217

            lr_kl[index17, index22] = lr2217
            lr_kl[index22, index17] = lr2217

            # second order mie between NH and H2O
            index14 = np.where(subgroups == 'H2O')[0]
            boolH2O = secondorder.group_l == 'H2O'
            eps2214 = secondorder[boolNH & boolH2O]['eps_kl'].values
            lr2214 = secondorder[boolNH & boolH2O]['lr_kl'].values

            eps_kl[index14, index22] = eps2214
            eps_kl[index22, index14] = eps2214
            lr_kl[index14, index22] = lr2214
            lr_kl[index22, index14] = lr2214


        del bool622, bool62, bool619

    bool22 = np.hstack(bool22)
    cond_asso22_0 = np.any(bool22)
    cond_asso22_14 = np.any(subgroup_id_asso == 'H2O')
    cond_asso22_17 = np.any(subgroup_id_asso == 'CO2')
    cond_asso22_19 = np.any(subgroup_id_asso == 'CH2OH')

    where22all = np.where(subgroup_id_asso == 'NH')
    where22all_id = molecule_id_index_asso[where22all]
    where22second = bool22[where22all_id]
    where22 = where22all[0][where22second]
    ncond22 = len(where22)

    if cond_asso22_0 and cond_asso22_14:
        where14 = np.where(subgroup_id_asso == 'H2O')
        boolk = secondasso.group_k == 'NH'
        booll = secondasso.group_l == 'H2O'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]
        for k in range(ncond22):
            for j in range(len1):
                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values
                if siteK == 'e1' and siteL == 'H':
                    indexH2O = sites_cumsum[where14]
                    indexNH = sites_cumsum[where22[k]] + 1
                    epsAB_kl[indexH2O, indexNH] = epsAB
                    epsAB_kl[indexNH, indexH2O] = epsAB
                    kAB_kl[indexH2O, indexNH] = kAB
                    kAB_kl[indexNH, indexH2O] = kAB
                elif siteK == 'H' and siteL == 'e1':
                    indexH2O = sites_cumsum[where14] + 1
                    indexNH = sites_cumsum[where22[k]]
                    epsAB_kl[indexH2O, indexNH] = epsAB
                    epsAB_kl[indexNH, indexH2O] = epsAB
                    kAB_kl[indexH2O, indexNH] = kAB
                    kAB_kl[indexNH, indexH2O] = kAB

    if cond_asso22_0 and cond_asso22_17:

        where17 = np.where(subgroup_id_asso == 'CO2')

        boolk = secondasso.group_k == 'NH'
        booll = secondasso.group_l == 'CO2'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]

        for k in range(ncond22):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values
                if siteK == 'e1' and siteL == 'e1':
                    indexCO2 = sites_cumsum[where17]
                    indexNH = sites_cumsum[where22[k]] + 1

                    epsAB_kl[indexCO2, indexNH] = epsAB
                    epsAB_kl[indexNH, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexNH] = kAB
                    kAB_kl[indexNH, indexCO2] = kAB

                elif siteK == 'e1' and siteL == 'e2':

                    indexCO2 = sites_cumsum[where17] + 1
                    indexNH = sites_cumsum[where22[k]] + 1

                    epsAB_kl[indexCO2, indexNH] = epsAB
                    epsAB_kl[indexNH, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexNH] = kAB
                    kAB_kl[indexNH, indexCO2] = kAB

    if cond_asso22_0 and cond_asso22_19:

        where19 = np.where(subgroup_id_asso == 'CH2OH')

        boolk = secondasso.group_k == 'NH'
        booll = secondasso.group_l == 'CH2OH'
        df = secondasso[boolk & booll]

        len1 = df.shape[0]

        for k in range(ncond22):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values
                if siteK == 'H' and siteL == 'e1':
                    indexOH = sites_cumsum[where19] + 1
                    indexNH = sites_cumsum[where22[k]]

                    epsAB_kl[indexOH, indexNH] = epsAB
                    epsAB_kl[indexNH, indexOH] = epsAB
                    kAB_kl[indexOH, indexNH] = kAB
                    kAB_kl[indexNH, indexOH] = kAB
                elif siteK == 'e1' and siteL == 'H':

                    indexOH = sites_cumsum[where19]
                    indexNH = sites_cumsum[where22[k]] + 1

                    epsAB_kl[indexOH, indexNH] = epsAB
                    epsAB_kl[indexNH, indexOH] = epsAB
                    kAB_kl[indexOH, indexNH] = kAB
                    kAB_kl[indexNH, indexOH] = kAB


def secondorder23(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                  secondorder, subgroup_id_asso, molecule_id_index_asso,
                  sites_cumsum, epsAB_kl, kAB_kl, secondasso):
    # condition between groups 'N'-'H2O' and 'N'-'CO2'
    bool23 = []
    for i in range(nc):
        index0 = group_indexes[i][0]
        indexf = group_indexes[i][1]
        subgroups_id_molecule = subgroups[index0:indexf]
        vki_molecule = vki[index0:indexf]

        cond7 = 'N' in subgroups_id_molecule and 'CH2' in subgroups_id_molecule and 'CH2OH' in subgroups_id_molecule and 'CH3' in subgroups_id_molecule
        cond70 = subgroups_id_molecule.shape[0] == 4
        cond71 = 4 <= np.sum(vki_molecule) <= 6

        bool723, bool719, bool72, bool73, bool71 = False, False, False, False, False

        if cond7 and cond70 and cond71:

            cond7_in23 = np.where(subgroups_id_molecule == 'N')
            bool723 = vki_molecule[cond7_in23] == 1

            cond7_in19 = np.where(subgroups_id_molecule == 'CH2OH')
            bool719 = vki_molecule[cond7_in19] in [1, 2]

            cond7_in2 = np.where(subgroups_id_molecule == 'CH2')
            bool72 = vki_molecule[cond7_in2] in [1, 2]

            bool73 = vki_molecule[cond7_in19] == vki_molecule[cond7_in2]

            cond7_in1 = np.where(subgroups_id_molecule == 'CH3')
            bool71 = vki_molecule[cond7_in1] in [1, 2]

        bool23.append(bool723 and bool719 and bool72 and bool73 and bool71)
        if bool723 and bool719 and bool72 and bool73 and bool71:

            index23 = cond7_in23[0] + index0

            index14 = np.where(subgroups == 'H2O')[0]

            boolN = secondorder.group_k == 'N'
            boolH2O = secondorder.group_l == 'H2O'
            eps2314 = secondorder[boolN & boolH2O]['eps_kl'].values
            lr2314 = secondorder[boolN & boolH2O]['lr_kl'].values

            index17 = np.where(subgroups == 'CO2')[0]

            boolN = secondorder.group_k == 'N'
            boolCO2 = secondorder.group_l == 'CO2'
            eps2317 = secondorder[boolN & boolCO2]['eps_kl'].values

            eps_kl[index23, index14] = eps2314
            eps_kl[index14, index23] = eps2314

            lr_kl[index14, index23] = lr2314
            lr_kl[index23, index14] = lr2314

            eps_kl[index17, index23] = eps2317
            eps_kl[index23, index17] = eps2317

        del bool723, bool719, bool72, bool73, bool71

    bool23 = np.hstack(bool23)
    cond_asso23_0 = np.any(bool23)
    cond_asso23_14 = np.any(subgroup_id_asso == 'H2O')
    cond_asso23_17 = np.any(subgroup_id_asso == 'CO2')
    cond_asso23_19 = np.any(subgroup_id_asso == 'CH2OH')

    where23all = np.where(subgroup_id_asso == 'N')
    where23all_id = molecule_id_index_asso[where23all]
    where23second = bool23[where23all_id]
    where23 = where23all[0][where23second]
    ncond23 = len(where23)

    if cond_asso23_0 and cond_asso23_14:

        where14 = np.where(subgroup_id_asso == 'H2O')

        boolk = secondasso.group_k == 'N'
        booll = secondasso.group_l == 'H2O'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]

        for k in range(ncond23):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values

                if siteK == 'e1' and siteL == 'H':

                    indexH = sites_cumsum[where14]
                    indexN = sites_cumsum[where23[k]]

                    epsAB_kl[indexH, indexN] = epsAB
                    epsAB_kl[indexN, indexH] = epsAB
                    kAB_kl[indexH, indexN] = kAB
                    kAB_kl[indexN, indexH] = kAB

    if cond_asso23_0 and cond_asso23_17:

        where17 = np.where(subgroup_id_asso == 'CO2')
        boolk = secondasso.group_k == 'N'
        booll = secondasso.group_l == 'CO2'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]

        for k in range(ncond23):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values

                if siteK == 'e1' and siteL == 'e1':
                    indexCO2 = sites_cumsum[where17]
                    indexN = sites_cumsum[where23[k]]

                    epsAB_kl[indexCO2, indexN] = epsAB
                    epsAB_kl[indexN, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexN] = kAB
                    kAB_kl[indexN, indexCO2] = kAB

                elif siteK == 'e1' and siteL == 'e2':

                    indexCO2 = sites_cumsum[where17] + 1
                    indexN = sites_cumsum[where23[k]]

                    epsAB_kl[indexCO2, indexN] = epsAB
                    epsAB_kl[indexN, indexCO2] = epsAB
                    kAB_kl[indexCO2, indexN] = kAB
                    kAB_kl[indexN, indexCO2] = kAB

    if cond_asso23_0 and cond_asso23_19:

        where19 = np.where(subgroup_id_asso == 'CH2OH')
        boolk = secondasso.group_k == 'N'
        booll = secondasso.group_l == 'CH2OH'
        df = secondasso[boolk & booll]
        len1 = df.shape[0]

        for k in range(ncond23):
            for j in range(len1):

                values = df.iloc[j].values[0:7]
                groupK2, siteK, groupL2, siteL, _, epsAB, kAB = values

                if siteK == 'e1' and siteL == 'H':

                    indexOH = sites_cumsum[where19]
                    indexN = sites_cumsum[where23[k]]

                    epsAB_kl[indexOH, indexN] = epsAB
                    epsAB_kl[indexN, indexOH] = epsAB
                    kAB_kl[indexOH, indexN] = kAB
                    kAB_kl[indexN, indexOH] = kAB



from __future__ import division, print_function, absolute_import
import numpy as np
from ..sgt import sgt_mix
from scipy.optimize import minimize_scalar


def fobj_beta(beta, iftexp, rho1, rho2, T, P, eos):
    bij = np.array([[0, beta], [beta, 0]])
    eos.beta_sgt(bij)
    tenb = np.zeros_like(iftexp)
    n = len(iftexp)

    n1, n2 = rho1.shape
    if n2 == n:
        rho1 = rho1.T
        rho2 = rho2.T

    for i in range(n):
        tenb[i] = sgt_mix(rho1[i], rho2[i], T[i], P[i], eos)
    fo = np.mean((1 - tenb/iftexp)**2)
    return fo


def fit_beta(beta0, ExpTension, EquilibriumInfo, eos, method='bounded', minimize_scalar_kwargs={}):
    """
    fit_beta
    Optimize beta for SGT for binary mixtures

    Parameters
    ----------
    beta0 : tuple
        boundaries for beta as needed for SciPy's minimize_scalar
    ExpTension : array
        Experimental interfacial tension of the mixture
    EquilibriumInfo : tuple
        tuple containing density vectors, temperature and pressure
        tuple = (rho1, rho2, T, P)
    eos : model
        saft vr mie model set up with the binary mixture
    method : str
        method to use for SciPy's minimize_scalar: 'bounded', 'brent' or 'golden' (default is 'bounded')
    minimize_scalar_kwargs : dict
        keyword arguments for SciPy's minimize_scalar

    Returns
    -------
    ten : OptimizeResult
        Result of SciPy minimize_scalar
    """
    rho1, rho2, T, P = EquilibriumInfo
    args = (ExpTension, rho1, rho2, T, P, eos)

    if method=='bounded':
        opti = minimize_scalar(fobj_beta, bounds=beta0, args=args, method=method, **minimize_scalar_kwargs)
    else:
        opti = minimize_scalar(fobj_beta, bracket=beta0, args=args, method=method, **minimize_scalar_kwargs)
    return opti



from __future__ import division, print_function, absolute_import
import numpy as np


# Third pertubation Eq 19
def a3m(x03, nsigma, eps3, f4, f5, f6):
    ter1 = -eps3*f4*nsigma
    ter2 = np.exp(f5*nsigma + f6*nsigma**2)
    return ter1*ter2


def da3m_deta(x03, nsigma, eps3, f4, f5, f6):

    aux = np.exp(f5*nsigma + f6*nsigma**2)
    a = -eps3*f4*nsigma
    a *= aux

    da = -eps3*f4*aux
    da *= (1. + nsigma * (f5 + 2*f6*nsigma))
    da *= x03
    return np.hstack([a, da])


def d2a3m_deta(x03, nsigma, eps3, f4, f5, f6):

    aux = np.exp(f5*nsigma + f6*nsigma**2)
    a = -eps3*f4*nsigma
    a *= aux

    da = -eps3*f4*aux
    da *= (1 + nsigma * (f5 + 2*f6*nsigma))
    da *= x03

    d2a = -eps3*f4*aux
    d2a *= (f5**2*nsigma+2*f6*nsigma*(3+2*f6*nsigma**2)+f5*(2+4*f6*nsigma**2))
    d2a *= x03**2
    return np.hstack([a, da, d2a])



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from .EquilibriumResult import EquilibriumResult
from ..constants import kb, Na


def fobj_crit(inc, eos):

    T, rho = inc

    h = 1e-4

    rho1 = rho + h
    rho2 = rho + 2*h

    P, dP = eos.dP_drho(rho, T)
    P1, dP1 = eos.dP_drho(rho1, T)
    P2, dP2 = eos.dP_drho(rho2, T)

    d2P_drho = (-1.5*dP + 2*dP1 - 0.5*dP2)/h

    # fo = np.array([-rho**2*dP, 2*rho**3*dP + rho**4*d2P_drho])
    # fo /= rho**2

    # fo = np.array([-dP, 2*dP + rho*d2P_drho])
    fo = np.array([-dP, 2*dP/rho + d2P_drho])

    return fo


def initial_guess_criticalpure(eos, n=50):
    """
    This functions computes inicial guesses of the critcal temperature in K
    and the critical density in mol/m^3.

    Function provided by Esteban Cea Klapp.

    Parameters
    ----------

    eos : Object
        saftvrmie pure object
    n: int
        lenght of linspace to study

    Returns
    -------
    Tc: float
        Critical temperature [K]
    rhoc: float
        Critical density [mol/m3]
    """

    romin = 0.7405
    romin *= 6/(Na*eos.xs_m*np.pi*eos.sigma_ii3)

    roc0 = romin/5
    Tc0 = eos.xs_m*eos.eps_ii/kb
    dTc0 = Tc0*0.3

    ro0 = 0.25*roc0
    rof = 2*roc0
    # Step 1. Check if tc0 is subcritical or supercritical
    ro = np.linspace(ro0, rof, n)
    dro = ro[1] - ro[0]
    for i in range(n):
        P, dP = eos.dP_drho(ro[i], Tc0)
        # If dP_drho < 0 in any interval, tc0 is subcritical (tc0 < tc)
        if dP < 0:
            # In this case you need to increase tc0
            dTc0 = Tc0*0.5
            T0 = 'sub'
            ro0 = ro[i] - dro
            break
        else:
            dTc0 = -Tc0*0.1
            T0 = 'sup'

    # Step 2. Find the temperature transition subcritical to supercritical
    loop = True
    ro = np.linspace(ro0, rof, n)
    dro = ro[1] - ro[0]
    k = 0
    T = [Tc0]  # List to store the studied temperatures
    if T0 == 'sub':
        while loop and k < 10:
            Tc0 += dTc0
            loop = False
            k += 1
            T.append(Tc0)
            for i in range(n):
                P, dP = eos.dP_drho(ro[i], Tc0)
                if dP < 0:
                    loop = True
                    ro0 = ro[i]
                    break
                loop = False
                Tsub = T[-2]
                Tsup = T[-1]
    else:
        while loop and k < 10:
            Tc0 += dTc0
            k += 1
            T.append(Tc0)
            for i in range(n):
                P, dP = eos.dP_drho(ro[i], Tc0)
                if dP < 0:
                    loop = False
                    ro0 = ro[i]
                    Tsup = T[-2]
                    Tsub = T[-1]
                    break

    # Step 3. Find rho_min and rho_max at the subcritical temperature

    ro = np.linspace(ro0, rof, n)
    dro = ro[1] - ro[0]
    ro_int = []
    flag = 0
    for i in range(n):
        P, dP = eos.dP_drho(ro[i], Tsub)
        if dP < 0:
            ro_int.append(ro[i])
            flag = 1
        else:
            if flag == 1:
                break

    # computing the initial guesses
    romin = min(ro_int)
    romax = max(ro_int)
    roc0 = (romin + romax)/2
    Tc0 = (Tsub + Tsup)/2

    return Tc0, roc0


def get_critical(eos, Tc0=None, rhoc0=None, method='hybr', full_output=False):
    """
    get_critical(eos, Tc0, rhoc0, method)

    Method that solves the critical coordinate of the fluid.

    Parameters
    ----------

    eos : Object
        saftvrmie pure object
    Tc0 : float
        initial guess for critical temperature [K]
    rhoc : float
        initial guess for critical density [mol/m^3]
    method : string, optional
        SciPy; root method to solve critical coordinate
    full_output: bool, optional
        whether to outputs or not all the calculation info.

    Returns
    -------
    Tc: float
        Critical temperature [K]
    Pc: float
        Critical pressure [Pa]
    rhoc: float
        Critical density [mol/m3]
    """

    if Tc0 is None and rhoc0 is None:
        Tc0, rhoc0 = initial_guess_criticalpure(eos, n=30)

    inc0 = np.array([Tc0, rhoc0])
    sol = root(fobj_crit, inc0, method=method, args=eos)
    Tc, rhoc = sol.x
    Pc = eos.pressure(rhoc, Tc)
    # print(sol)
    if full_output:
        dict = {'Tc': Tc, 'Pc': Pc, 'rhoc': rhoc, 'error': sol.fun,
                'nfev:': sol.nfev, 'message': sol.message,
                'success': sol.success}
        out = EquilibriumResult(dict)
    else:
        out = Tc, Pc, rhoc
    return out



from __future__ import division, print_function, absolute_import
import numpy as np


# Eq (24) Paper 2014
def J_lam(x0, lam):
    lam3 = 3. - lam
    lam4 = 4. - lam
    J = (lam3*x0**lam4 - lam4*x0**lam3 + 1.)/(lam3 * lam4)
    return J


# Eq (23) Paper 2014
def I_lam(x0, lam):
    lam3 = 3. - lam
    I = (x0**lam3-1.) / lam3
    return I


# Equation (26) Paper 2014
def xhi_eff(xhix_vec, cictes):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([xhix, xhix2, xhix3, xhix4])
    xhieff = np.matmul(cictes, xhix_vec)
    return xhieff


def dxhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3*xhix2, 4*xhix3]])

    xhieff = np.matmul(cictes, xhix_vec[0])
    dxhieff = np.matmul(cictes, xhix_vec[1])
    dxhieff *= dxhix_dxhi00
    return xhieff, dxhieff


def d2xhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3*xhix2, 4*xhix3],
    #                      [0., 2., 6*xhix, 12*xhix2]])

    xhieff = np.matmul(cictes, xhix_vec[0])
    dxhieff = np.matmul(cictes, xhix_vec[1])
    dxhieff *= dxhix_dxhi00
    d2xhieff = np.matmul(cictes, xhix_vec[2])
    d2xhieff *= dxhix_dxhi00**2
    return xhieff, dxhieff, d2xhieff


def d3xhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3.*xhix2, 4.*xhix3],
    #                      [0., 2., 6*xhix, 12.*xhix2],
    #                     [0., 0., 6., 24.*xhix]])
    dxhix_dxhi00_2 = dxhix_dxhi00**2
    dxhix_dxhi00_3 = dxhix_dxhi00_2*dxhix_dxhi00

    xhieff = np.matmul(cictes, xhix_vec[0])
    dxhieff = np.matmul(cictes, xhix_vec[1])
    dxhieff *= dxhix_dxhi00
    d2xhieff = np.matmul(cictes, xhix_vec[2])
    d2xhieff *= dxhix_dxhi00_2
    d3xhieff = np.matmul(cictes, xhix_vec[3])
    d3xhieff *= dxhix_dxhi00_3
    return xhieff, dxhieff, d2xhieff, d3xhieff


def dxhieff_dx(xhix_vec, cictes, dxhix_dx):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3*xhix2, 4*xhix3]])
    xhieff = np.matmul(cictes, xhix_vec[0])
    dxhieff = np.multiply.outer(dxhix_dx, np.matmul(cictes, xhix_vec[1]))
    return xhieff, dxhieff


def dxhieff_dx_dxhi00(xhix_vec, cictes,  dxhix_dxhi00, dxhix_dx):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3*xhix2, 4*xhix3]])

    xhieff = np.matmul(cictes, xhix_vec[0])
    aux1 = np.matmul(cictes, xhix_vec[1])
    dxhieff = aux1 * dxhix_dxhi00
    dxhieff_dx = np.multiply.outer(dxhix_dx, aux1)
    return xhieff, dxhieff, dxhieff_dx


def dxhieff_dx_dxhi00_dxxhi(xhix_vec, cictes, dxhix_dxhi00, dxhix_dx,
                            dxhix_dx_dxhi00):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                     [1., 2 * xhix, 3*xhix2, 4*xhix3],
    #                     [0., 2, 6*xhix, 12*xhix2]])

    xhieff = np.matmul(cictes, xhix_vec[0])
    aux_der1 = np.matmul(cictes, xhix_vec[1])
    aux_der2 = np.matmul(cictes, xhix_vec[2])

    dxhieff = dxhix_dxhi00 * aux_der1
    dxhieff_dx = np.multiply.outer(dxhix_dx, aux_der1)

    aux2 = np.multiply.outer(dxhix_dx_dxhi00, aux_der1)
    aux3 = np.multiply.outer(dxhix_dx, aux_der2)
    dxhieff_dxxhi = aux2 + dxhix_dxhi00 * aux3

    return xhieff, dxhieff, dxhieff_dx, dxhieff_dxxhi


def dxhieff_dx_d2xhi00_dxxhi(xhix_vec, cictes, dxhix_dxhi00, dxhix_dx,
                             dxhix_dx_dxhi00):
    # xhix2 = xhix**2
    # xhix3 = xhix2*xhix
    # xhix4 = xhix3*xhix
    # xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
    #                      [1., 2 * xhix, 3*xhix2, 4*xhix3],
    #                     [0., 2, 6*xhix, 12*xhix2]])

    xhieff = np.matmul(cictes, xhix_vec[0])
    aux_der1 = np.matmul(cictes, xhix_vec[1])
    aux_der2 = np.matmul(cictes, xhix_vec[2])

    dxhieff = dxhix_dxhi00 * aux_der1
    d2xhieff = dxhix_dxhi00**2 * aux_der2
    dxhieff_dx = np.multiply.outer(dxhix_dx, aux_der1)

    aux2 = np.multiply.outer(dxhix_dx_dxhi00, aux_der1)
    aux3 = np.multiply.outer(dxhix_dx, aux_der2)
    dxhieff_dxxhi = aux2 + dxhix_dxhi00 * aux3

    return xhieff, dxhieff, d2xhieff, dxhieff_dx, dxhieff_dxxhi


# Equation (32) Paper 2014
def Xi(xhixm, f1, f2, f3):
    x = f1*xhixm + f2*xhixm**5 + f3*xhixm**8
    return x


def dXi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3):
    xhixm2 = xhixm**2
    xhixm4 = xhixm2**2
    xhixm5 = xhixm4*xhixm
    xhixm7 = xhixm5*xhixm2
    xhixm8 = xhixm7*xhixm

    x = f1*xhixm + f2*xhixm5 + f3*xhixm8

    dx = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dx *= dxhixm_dxhi00
    return x, dx


def d2Xi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3):
    xhixm2 = xhixm**2
    xhixm3 = xhixm2*xhixm
    xhixm4 = xhixm3*xhixm
    xhixm5 = xhixm4*xhixm
    xhixm6 = xhixm5*xhixm
    xhixm7 = xhixm6*xhixm
    xhixm8 = xhixm7*xhixm

    x = f1*xhixm + f2*xhixm5 + f3*xhixm8

    dx = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dx *= dxhixm_dxhi00

    d2x = 20*f2*xhixm3 + 56*f3*xhixm6
    d2x *= dxhixm_dxhi00**2
    return x, dx, d2x


def dXi_dx(xhixm, dxhixm_dx, f1, f2, f3):
    xhixm4 = xhixm**4
    xhixm5 = xhixm4*xhixm
    xhixm7 = xhixm5*xhixm**2
    xhixm8 = xhixm7*xhixm

    xi = f1*xhixm + f2*xhixm5 + f3*xhixm8

    dxi_aux = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dxi = np.multiply.outer(dxhixm_dx, dxi_aux)
    return xi, dxi


def dXi_dxxhi(xhixm, dxhixm_dxhi00, dxhixm_dx, f1, f2, f3):
    xhixm4 = xhixm**4
    xhixm5 = xhixm4*xhixm
    xhixm7 = xhixm5*xhixm**2
    xhixm8 = xhixm7*xhixm

    xi = f1*xhixm + f2*xhixm5 + f3*xhixm8

    aux_dxi = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dxi_dxhi = aux_dxi * dxhixm_dxhi00

    dxi_dx = np.multiply.outer(dxhixm_dx, aux_dxi)
    return xi, dxi_dxhi, dxi_dx


# Equation (31) Paper 2014
def kHS(xhix):
    k = (1-xhix)**4
    k /= 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4
    return k


def dkHS_dxhi00(xhix, dxhix_dxhi00):
    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix3 * xhix

    den = 1 + 4*xhix + 4*xhix2 - 4*xhix3 + xhix4
    den2 = den**2

    xhix_1 = -1 + xhix
    xhix_12 = xhix_1**2
    xhix_13 = xhix_12*xhix_1
    xhix_14 = xhix_13*xhix_1

    k = xhix_14/den

    dk = -2 - 5*xhix + xhix2
    dk *= -4 * xhix_13 * dxhix_dxhi00
    dk /= den2
    return k, dk


def d2kHS_dxhi00(xhix, dxhix_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix3 * xhix
    xhix5 = xhix4 * xhix
    xhix6 = xhix5 * xhix

    den = 1 + 4*xhix + 4*xhix2 - 4*xhix3 + xhix4
    den2 = den**2
    den3 = den2*den

    xhix_1 = -1 + xhix
    xhix_12 = xhix_1**2
    xhix_13 = xhix_12*xhix_1
    xhix_14 = xhix_13*xhix_1

    k = xhix_14/den

    dk = -2 - 5*xhix + xhix2
    dk *= -4 * xhix_13 * dxhix_dxhi00
    dk /= den2

    d2k = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    d2k += 77 * xhix4 - 30 * xhix5 + 3*xhix6
    d2k *= 4 * xhix_12 * dxhix_dxhi00**2
    d2k /= den3

    return k, dk, d2k


def d3kHS_dxhi00(xhix, dxhix_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix3 * xhix
    xhix5 = xhix4 * xhix
    xhix6 = xhix5 * xhix
    xhix7 = xhix6 * xhix
    xhix8 = xhix7 * xhix
    xhix9 = xhix8 * xhix

    den = 1 + 4*xhix + 4*xhix2 - 4*xhix3 + xhix4
    den2 = den**2
    den3 = den2*den
    den4 = den3*den

    xhix_1 = -1 + xhix
    xhix_12 = xhix_1**2
    xhix_13 = xhix_12*xhix_1
    xhix_14 = xhix_13*xhix_1

    k = xhix_14/den

    dk = -2 - 5*xhix + xhix2
    dk *= -4 * xhix_13 * dxhix_dxhi00
    dk /= den2

    d2k = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    d2k += 77 * xhix4 - 30 * xhix5 + 3*xhix6
    d2k *= 4 * xhix_12 * dxhix_dxhi00**2
    d2k /= den3

    d3k = -97 - 109 * xhix + 238 * xhix2 - 352 * xhix4
    d3k += 372 * xhix5 - 210*xhix6 + 77*xhix7 - 15*xhix8 + xhix9
    d3k *= xhix
    d3k -= 13.
    d3k *= 12.*(-4 * xhix_1 * dxhix_dxhi00**3)
    d3k /= den4

    return k, dk, d2k, d3k


def dkHS_dx(xhix, dxhix_dx):
    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dx
    dk /= (1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4)**2
    return dk


def dkHS_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix5 = xhix4*xhix
    xhix6 = xhix5*xhix

    den = 1 + 4*xhix + 4*xhix2 - 4*xhix3 + xhix4
    den2 = den**2
    den3 = den2*den

    xhix_1 = -1 + xhix
    xhix_12 = xhix_1**2
    xhix_13 = xhix_12*xhix_1
    xhix_14 = xhix_13*xhix_1

    k = xhix_14/den

    auxdk = -2 - 5*xhix + xhix2
    auxdk *= -4 * xhix_13
    auxdk /= den2

    dk = auxdk*dxhix_dxhi00
    dkx = auxdk*dxhix_dx

    aux1 = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    aux1 += 77 * xhix4 - 30 * xhix5 + 3*xhix6

    aux2 = -2 - 13 * xhix - 27*xhix2 - 8 * xhix3
    aux2 += 22 * xhix4 - 9 * xhix5 + xhix6

    dkxxhi = dxhix_dxhi00 * dxhix_dx * aux1
    dkxxhi += -xhix_1*aux2*dxhix_dx_dxhi00

    dkxxhi *= 4 * xhix_12
    dkxxhi /= den3
    return k, dk, dkx, dkxxhi


def d2kHS_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix5 = xhix4*xhix
    xhix6 = xhix5*xhix

    den = 1 + 4*xhix + 4*xhix2 - 4*xhix3 + xhix4
    den2 = den**2
    den3 = den2*den

    xhix_1 = -1 + xhix
    xhix_12 = xhix_1**2
    xhix_13 = xhix_12*xhix_1
    xhix_14 = xhix_13*xhix_1

    k = xhix_14/den

    auxdk = -2 - 5*xhix + xhix2
    auxdk *= -4 * xhix_13
    auxdk /= den2

    dk = auxdk*dxhix_dxhi00
    dkx = auxdk*dxhix_dx

    aux1 = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    aux1 += 77 * xhix4 - 30 * xhix5 + 3*xhix6

    d2k = 4 * xhix_12 * dxhix_dxhi00**2 * aux1
    d2k /= den3

    aux2 = -2 - 13 * xhix - 27*xhix2 - 8 * xhix3
    aux2 += 22 * xhix4 - 9 * xhix5 + xhix6

    dkxxhi = dxhix_dxhi00 * dxhix_dx * aux1
    dkxxhi += -xhix_1*aux2*dxhix_dx_dxhi00

    dkxxhi *= 4 * xhix_12
    dkxxhi /= den3

    return k, dk, d2k, dkx, dkxxhi



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import fsolve


def association_config(sitesmix, eos):

    compindex = []
    for i, j in enumerate(sitesmix):
        compindex += np.count_nonzero(j) * [i]
    compindex = np.asarray(compindex)

    types = np.array([['B', 'P', 'N']])
    n = len(sitesmix)
    n = eos.nc
    types = np.repeat(types, n, axis=0)
    nozero = np.nonzero(sitesmix)
    types = types[nozero]
    ntypes = np.asarray(sitesmix)[nozero]
    nsites = len(types)

    S = np.array(sitesmix)
    S = S[S != 0]

    DIJ = np.zeros([nsites, nsites])
    int_i = []
    int_j = []
    for i in range(nsites):
        for j in range(nsites):
            bool1 = types[i] == 'B'
            bool2 = types[i] == 'P' and (types[j] == 'N' or types[j] == 'B')
            bool3 = types[i] == 'N' and (types[j] == 'P' or types[j] == 'B')
            if bool1 or bool2 or bool3:
                DIJ[i, j] = ntypes[j]
                int_i.append(i)
                int_j.append(j)

    indexabij = (int_i, int_j)
    indexab = (compindex[int_i], compindex[int_j])
    diagasso = np.diag_indices(nsites)

    dxjdx = np.zeros([eos.nc, nsites])
    for i in range(eos.nc):
        dxjdx[i] = compindex == i

    return S, DIJ, compindex, indexabij, indexab, nsites, dxjdx, diagasso


def Iab(xhi, aux_dii, aux_dii2, Kab):

    xhi0, xhi1, xhi2, xhi3 = xhi
    # aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
    aux = aux_dii
    aux_2 = aux_dii2

    xhi13 = 1 - xhi3
    xhi13_2 = xhi13**2
    xhi13_3 = xhi13_2*xhi13

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13_2
    gdhs += 2 * aux_2 * xhi2**2 / xhi13_3

    iab = gdhs * Kab
    return iab


def dIab_drho(xhi, dxhi_dxhi00, dxhi00_drho, aux_dii, aux_dii2, Kab):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    # aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
    # aux_2 = aux**2
    aux = aux_dii
    aux_2 = aux_dii2

    xhi2_2 = xhi2**2

    xhi13 = 1 - xhi3
    xhi13_2 = xhi13**2
    xhi13_3 = xhi13_2*xhi13
    xhi13_4 = xhi13_3*xhi13

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13_2
    gdhs += 2 * aux_2 * xhi2_2 / xhi13_3

    dgdhs = 4. * aux_2 * xhi2 * dxhi2 / xhi13_3
    dgdhs += 3. * aux * dxhi2 / xhi13_2
    dgdhs += 6. * aux_2 * xhi2_2 * dxhi3 / xhi13_4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13_3
    dgdhs += dxhi3 / xhi13_2
    dgdhs *= dxhi00_drho

    iab = gdhs * Kab
    diab = dgdhs * Kab
    return iab, diab


def d2Iab_drho(xhi, dxhi_dxhi00, dxhi00_drho, aux_dii, aux_dii2, Kab):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    # aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
    # aux_2 = aux**2
    aux = aux_dii
    aux_2 = aux_dii2

    xhi2_2 = xhi2**2
    dxhi3_2 = dxhi3**2
    xhi13 = 1 - xhi3
    xhi13_2 = xhi13**2
    xhi13_3 = xhi13_2*xhi13
    xhi13_4 = xhi13_3*xhi13
    xhi13_5 = xhi13_4*xhi13

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13_2
    gdhs += 2 * aux_2 * xhi2_2 / xhi13_3

    dgdhs = 4. * aux_2 * xhi2 * dxhi2 / xhi13_3
    dgdhs += 3. * aux * dxhi2 / xhi13_2
    dgdhs += 6. * aux_2 * xhi2_2 * dxhi3 / xhi13_4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13_3
    dgdhs += dxhi3 / xhi13_2
    dgdhs *= dxhi00_drho

    d2gdhs = 2 * dxhi3_2 / xhi13_3
    d2gdhs += 2 * aux_2 * 2 * dxhi2**2 / xhi13_3
    d2gdhs += 2 * aux_2 * 12 * xhi2 * dxhi2 * dxhi3 / xhi13_4
    d2gdhs += 2 * aux_2 * 12 * xhi2_2 * dxhi3_2 / xhi13_5
    d2gdhs += 3 * aux * 4 * dxhi2 * dxhi3 / xhi13_3
    d2gdhs += 3 * aux * 6 * xhi2 * dxhi3_2 / xhi13_4
    d2gdhs *= dxhi00_drho**2

    iab = gdhs * Kab
    diab = dgdhs * Kab
    d2iab = d2gdhs * Kab

    return iab, diab, d2iab


def dIab_dx(xhi, dxhi_dx, aux_dii, aux_dii2, Kab):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx

    # aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
    # aux_2 = aux**2
    aux = aux_dii
    aux_2 = aux_dii2
    xhi2_2 = xhi2**2

    xhi13 = 1 - xhi3
    xhi13_2 = xhi13**2
    xhi13_3 = xhi13_2*xhi13
    xhi13_4 = xhi13_3*xhi13

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13_2
    gdhs += 2 * aux_2 * xhi2_2 / xhi13_3

    dgdhsx = 4 * np.multiply.outer(aux_2, dxhi2x) * xhi2 / xhi13_3
    dgdhsx += 3 * np.multiply.outer(aux, dxhi2x) / xhi13_2
    dgdhsx += 6 * np.multiply.outer(aux_2, dxhi3x) * xhi2_2 / xhi13_4
    dgdhsx += 6 * np.multiply.outer(aux, dxhi3x) * xhi2 / xhi13_3
    dgdhsx += dxhi3x / xhi13_2
    dgdhsx = dgdhsx.T

    iab = gdhs * Kab
    diabx = dgdhsx * Kab
    return iab, diabx


def dIab_dxrho(xhi, dxhi_dxhi00, dxhi00_drho, dxhi_dx, aux_dii, aux_dii2, Kab):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx
    # aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
    # aux_2 = aux**2
    aux = aux_dii
    aux_2 = aux_dii2
    xhi2_2 = xhi2**2

    xhi13 = 1 - xhi3
    xhi13_2 = xhi13**2
    xhi13_3 = xhi13_2*xhi13
    xhi13_4 = xhi13_3*xhi13

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13_2
    gdhs += 2 * aux_2 * xhi2_2 / xhi13_3

    dgdhs = 4. * aux_2 * xhi2 * dxhi2 / xhi13_3
    dgdhs += 3. * aux * dxhi2 / xhi13_2
    dgdhs += 6. * aux_2 * xhi2_2 * dxhi3 / xhi13_4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13_3
    dgdhs += dxhi3 / xhi13_2
    dgdhs *= dxhi00_drho

    dgdhsx = 4 * np.multiply.outer(aux_2, dxhi2x) * xhi2 / xhi13_3
    dgdhsx += 3 * np.multiply.outer(aux, dxhi2x) / xhi13_2
    dgdhsx += 6 * np.multiply.outer(aux_2, dxhi3x) * xhi2_2 / xhi13_4
    dgdhsx += 6 * np.multiply.outer(aux, dxhi3x) * xhi2 / xhi13_3
    dgdhsx += dxhi3x / xhi13_2
    dgdhsx = dgdhsx.T

    iab = gdhs * Kab
    diab = dgdhs * Kab
    diabx = dgdhsx * Kab
    return iab, diab, diabx


def fobj_xass(Xass, xj, aux_asso, diagasso):
    fo = Xass - 1 / (1 + aux_asso@(xj*Xass))
    return fo


def fobj_xass_jac(Xass, xj, aux_asso, diagasso):
    den = 1 + aux_asso@(xj*Xass)
    dfo = ((aux_asso*xj).T/den**2).T
    dfo[diagasso] += 1.
    return dfo


def Xass_solver(nsites, xj, rho, DIJ, Dabij, diagasso, Xass0):

    Xass = 1. * Xass0

    aux_asso = rho * DIJ * Dabij

    omega = 0.2
    for i in range(5):
        den = 1. + aux_asso@(xj*Xass)
        fo = 1. / den
        dXass = (1 - omega) * (fo - Xass)
        Xass += dXass

    bool_method = not np.any(xj == 0.)

    if bool_method:
        KIJ = np.outer(xj, xj) * aux_asso

        KIJXass = KIJ@Xass
        dQ = xj * (1/Xass - 1) - KIJXass
        HIJ = -1 * KIJ
        HIJ[diagasso] -= (xj + KIJXass)/Xass
        for i in range(15):
            dXass = np.linalg.solve(HIJ, -dQ)
            Xnew = Xass + dXass

            is_nan = np.isnan(Xnew)
            Xnew[is_nan] = 0.2

            Xnew_neg = Xnew < 0.
            Xnew[Xnew_neg] = 0.2*Xass[Xnew_neg]

            Xnew_big = Xnew > 1.
            Xnew[Xnew_big] = 1.

            Xass = Xnew
            KIJXass = KIJ@Xass
            dQ = xj * (1/Xass - 1) - KIJXass
            sucess = np.linalg.norm(dQ) < 1e-9
            if sucess:
                break
            HIJ = -1 * KIJ
            HIJ[diagasso] -= (xj + KIJXass)/Xass
    else:
        Xass = fsolve(fobj_xass, x0=Xass, args=(xj, aux_asso, diagasso),
                      fprime=fobj_xass_jac)

    return Xass


def CIJ_matrix(rhom, xj, Xass, DIJ, Dabij, diagasso):
    CIJ = rhom * np.outer(Xass**2, xj) * Dabij * DIJ
    CIJ[diagasso] += 1.
    return CIJ


def dXass_drho(rhom, xj, Xass, DIJ, Dabij, dDabij_drho, CIJ):
    brho = -(DIJ*(Dabij + rhom * dDabij_drho))@(xj * Xass)
    brho *= Xass**2
    dXass = np.linalg.solve(CIJ, brho)
    return dXass


def d2Xass_drho(rhom, xj, Xass, dXass_drho, DIJ, Dabij, dDabij_drho,
                d2Dabij_drho, CIJ):

    b2rho = -rhom * (DIJ * (Xass * d2Dabij_drho+2*dXass_drho*dDabij_drho))@xj
    b2rho += 2 * (1/Xass - 1) / (rhom**2)
    b2rho *= Xass**2
    b2rho += 2 * dXass_drho / (rhom)
    b2rho += 2 * dXass_drho**2 / (Xass)

    d2Xass = np.linalg.solve(CIJ, b2rho)
    return d2Xass


def dXass_dx(rhom, xj, Xass, DIJ, Dabij, dDabij_dx, dxjdx, CIJ):

    bx = (dxjdx * Xass)@(DIJ*Dabij).T + (DIJ * dDabij_dx)@(xj * Xass)
    bx *= Xass**2
    bx *= -rhom
    dXass = np.linalg.solve(CIJ, bx.T)

    return dXass.T


def association_solver(self, x, rhom, temp_aux, Xass0=None):

    if Xass0 is None:
        Xass = 0.2 * np.ones(self.nsites)
    else:
        Xass = 1. * Xass0

    di03 = temp_aux[8]
    dxhi00_drho = self.dxhi00_drho
    xmi = x * self.ms
    xm = np.sum(xmi)
    # Eq A8
    xs = xmi / xm

    # Defining xhi0 without x dependence
    xhi00 = dxhi00_drho * rhom
    # Eq A7
    xhi = xhi00 * xm * np.matmul(xs, di03)
    dxhi_dxhi00 = np.matmul(xmi, di03)
    dxhi_dxhi00[0] = xm

    xj = x[self.compindex]
    Fab = temp_aux[25]
    aux_dii = temp_aux[26]
    aux_dii2 = temp_aux[27]
    Kab = temp_aux[28]
    iab = Iab(xhi, aux_dii, aux_dii2, Kab)
    Dab = self.sigmaij3 * Fab * iab
    Dabij = np.zeros([self.nsites, self.nsites])
    Dabij[self.indexabij] = Dab[self.indexab]

    Xass = Xass_solver(self.nsites, xj, rhom, self.DIJ, Dabij,
                       self.diagasso, Xass)
    return Xass


def association_check(self, x, rhom, temp_aux, Xass):
    di03 = temp_aux[8]
    dxhi00_drho = self.dxhi00_drho
    xmi = x * self.ms
    xm = np.sum(xmi)
    # Eq A8
    xs = xmi / xm

    # Defining xhi0 without x dependence
    xhi00 = dxhi00_drho * rhom
    # Eq A7
    xhi = xhi00 * xm * np.matmul(xs, di03)
    dxhi_dxhi00 = np.matmul(xmi, di03)
    dxhi_dxhi00[0] = xm

    xj = x[self.compindex]
    Fab = temp_aux[25]
    aux_dii = temp_aux[26]
    aux_dii2 = temp_aux[27]
    Kab = temp_aux[28]
    iab = Iab(xhi, aux_dii, aux_dii2, Kab)
    Dab = self.sigmaij3 * Fab * iab
    Dabij = np.zeros([self.nsites, self.nsites])
    Dabij[self.indexabij] = Dab[self.indexab]

    aux_asso = rhom * self.DIJ * Dabij
    fo = Xass - 1 / (1 + aux_asso@(xj*Xass))
    return fo



from __future__ import division, print_function, absolute_import
import numpy as np
from .B_monomer import B, dB_dxhi00, d2B_dxhi00, d3B_dxhi00
from .B_monomer import dB_dx, dB_dx_dxhi00, dB_dx_dxhi00_dxxhi
from .B_monomer import dB_dx_d2xhi00_dxxhi
from .a1s_monomer import a1s, da1s_dxhi00, d2a1s_dxhi00, d3a1s_dxhi00
from .a1s_monomer import da1s_dx, da1s_dx_dxhi00, da1s_dx_dxhi00_dxxhi
from .a1s_monomer import da1s_dx_d2xhi00_dxxhi


def a1sB(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw, a1vdw_cte):
    a1 = a1s(xhi00, xhix_vec, xm, cictes, a1vdw)
    b = B(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte)
    return a1 + b


def da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw, a1vdw_cte,
                 dxhix_dxhi00):
    a1, da1 = da1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                          dxhix_dxhi00)
    b, db = dB_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00)
    return a1 + b, da1 + db


def d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw,
                  a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1 = d2a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                                 dxhix_dxhi00)
    b, db, d2b = d2B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b


def d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw,
                  a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1, d3a1 = d3a1s_dxhi00(xhi00, xhix_vec, xm, cictes,
                                       a1vdw, dxhix_dxhi00)
    b, db, d2b, d3b = d3B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte,
                                 dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b, d3a1 + d3b


def da1sB_dx(xhi00, xhix, xhix_vec, xm, ms, Ikl, Jkl, cictes, a1vdw, a1vdw_cte,
             dxhix_dx):
    a1, da1x = da1s_dx(xhi00, xhix_vec, xm, ms, cictes, a1vdw, dxhix_dx)
    b, dbx = dB_dx(xhi00, xhix, xm, ms, Ikl, Jkl, a1vdw_cte,  dxhix_dx)
    return a1 + b, da1x + dbx


def da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xm, ms, Ikl, Jkl, cictes, a1vdw,
                    a1vdw_cte, dxhix_dxhi00, dxhix_dx):
    a1, da1, da1x = da1s_dx_dxhi00(xhi00, xhix_vec, xm, ms, cictes,
                                   a1vdw, dxhix_dxhi00, dxhix_dx)
    b, db, dbx = dB_dx_dxhi00(xhi00, xhix, xm, ms, Ikl, Jkl, a1vdw_cte,
                              dxhix_dxhi00, dxhix_dx)
    return a1+b, da1+db, da1x+dbx


def da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, Ikl, Jkl, cictes,
                          a1vdw, a1vdw_cte, dxhix_dxhi00, dxhix_dx,
                          dxhix_dx_dxhi00):
    out = da1s_dx_dxhi00_dxxhi(xhi00, xhix_vec, xm, ms, cictes, a1vdw,
                               dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1, da1, da1x, da1xxhi = out
    out = dB_dx_dxhi00_dxxhi(xhi00, xhix, xm, ms, Ikl, Jkl, a1vdw_cte,
                             dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    b, db, dbx, dbxxhi = out
    return a1+b, da1+db, da1x+dbx, da1xxhi+dbxxhi


def da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, Ikl, Jkl, cictes,
                           a1vdw, a1vdw_cte, dxhix_dxhi00, dxhix_dx,
                           dxhix_dx_dxhi00):
    out = da1s_dx_d2xhi00_dxxhi(xhi00, xhix_vec, xm, ms, cictes, a1vdw,
                                dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1, da1, d2a1, da1x, da1xxhi = out
    out = dB_dx_d2xhi00_dxxhi(xhi00, xhix, xm, ms, Ikl, Jkl, a1vdw_cte,
                              dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    b, db, d2b, dbx, dbxxhi = out
    return a1+b, da1+db, d2a1+d2b, da1x+dbx, da1xxhi+dbxxhi


def a1sB_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl, ccteskl,
              a1vdwkl, a1vdw_ctekl):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    a1sb_a = a1sB(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl, cctes_lakl,
                  a1vdw_lakl, a1vdw_ctekl)
    a1sb_r = a1sB(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl, cctes_lrkl,
                  a1vdw_lrkl, a1vdw_ctekl)
    a1sb_2a = a1sB(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl, cctes_2lakl,
                   a1vdw_2lakl, a1vdw_ctekl)
    a1sb_2r = a1sB(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl, cctes_2lrkl,
                   a1vdw_2lrkl, a1vdw_ctekl)
    a1sb_ar = a1sB(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl, cctes_larkl,
                   a1vdw_larkl, a1vdw_ctekl)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])
    return a1sb_a1, a1sb_a2


def da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                      ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    a1sb_a, da1sb_a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                                   cctes_lakl, a1vdw_lakl, a1vdw_ctekl,
                                   dxhix_dxhi00)

    a1sb_r, da1sb_r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                                   cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl,
                                   dxhix_dxhi00)

    a1sb_2a, da1sb_2a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl,
                                     J_2lakl, cctes_2lakl, a1vdw_2lakl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_2r, da1sb_2r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl,
                                     J_2lrkl, cctes_2lrkl, a1vdw_2lrkl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_ar, da1sb_ar = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl,
                                     J_larkl, cctes_larkl, a1vdw_larkl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])
    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    return a1sb_a1, a1sb_a2


def d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                       ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                           cctes_lakl, a1vdw_lakl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a = out_la

    out_lr = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                           cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r = out_lr

    out_2la = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl,
                            cctes_2lakl, a1vdw_2lakl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a = out_2la

    out_2lr = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl,
                            cctes_2lrkl, a1vdw_2lrkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r = out_2lr

    out_lar = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl,
                            cctes_larkl, a1vdw_larkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])
    return a1sb_a1, a1sb_a2


def d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                       ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                           cctes_lakl, a1vdw_lakl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a, d3a1sb_a = out_la

    out_lr = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                           cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r, d3a1sb_r = out_lr

    out_2la = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl,
                            cctes_2lakl, a1vdw_2lakl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a, d3a1sb_2a = out_2la

    out_2lr = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl,
                            cctes_2lrkl, a1vdw_2lrkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r, d3a1sb_2r = out_2lr

    out_lar = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl,
                            cctes_larkl, a1vdw_larkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar, d3a1sb_ar = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r],
                       [d3a1sb_a, d3a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r],
                       [d3a1sb_2a, d3a1sb_ar, d3a1sb_2r]])

    return a1sb_a1, a1sb_a2


def da1sB_dx_eval(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lambdaskl, J_lambdaskl,
                  ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dx):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    a1sb_a, da1sb_a = da1sB_dx(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lakl,
                               J_lakl, cctes_lakl, a1vdw_lakl, a1vdw_ctekl,
                               dxhix_dx)
    a1sb_r, da1sb_r = da1sB_dx(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lrkl,
                               J_lrkl, cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl,
                               dxhix_dx)
    a1sb_2a, da1sb_2a = da1sB_dx(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lakl,
                                 J_2lakl, cctes_2lakl, a1vdw_2lakl,
                                 a1vdw_ctekl, dxhix_dx)
    a1sb_2r, da1sb_2r = da1sB_dx(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lrkl,
                                 J_2lrkl, cctes_2lrkl, a1vdw_2lrkl,
                                 a1vdw_ctekl, dxhix_dx)
    a1sb_ar, da1sb_ar = da1sB_dx(xhi00, xhix, xhix_vec, xs_m, zs_m, I_larkl,
                                 J_larkl, cctes_larkl, a1vdw_larkl,
                                 a1vdw_ctekl, dxhix_dx)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])

    a1sb_a1x = np.array([da1sb_a, da1sb_r])
    a1sb_a2x = np.array([da1sb_2a, da1sb_ar, da1sb_2r])
    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x


def da1sB_dx_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lambdaskl,
                         J_lambdaskl, ccteskl, a1vdwkl, a1vdw_ctekl,
                         dxhix_dxhi00, dxhix_dx):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lakl, J_lakl,
                             cctes_lakl, a1vdw_lakl, a1vdw_ctekl, dxhix_dxhi00,
                             dxhix_dx)
    a1sb_a, da1sb_a, da1sb_ax = out_la
    out_lr = da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lrkl, J_lrkl,
                             cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl, dxhix_dxhi00,
                             dxhix_dx)
    a1sb_r, da1sb_r, da1sb_rx = out_lr
    out_2la = da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lakl,
                              J_2lakl, cctes_2lakl, a1vdw_2lakl, a1vdw_ctekl,
                              dxhix_dxhi00, dxhix_dx)
    a1sb_2a, da1sb_2a, da1sb_2ax = out_2la
    out_2lr = da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lrkl,
                              J_2lrkl, cctes_2lrkl, a1vdw_2lrkl, a1vdw_ctekl,
                              dxhix_dxhi00, dxhix_dx)
    a1sb_2r, da1sb_2r, da1sb_2rx = out_2lr
    out_lar = da1sB_dx_dxhi00(xhi00, xhix, xhix_vec, xs_m, zs_m, I_larkl,
                              J_larkl, cctes_larkl, a1vdw_larkl, a1vdw_ctekl,
                              dxhix_dxhi00, dxhix_dx)
    a1sb_ar, da1sb_ar, da1sb_arx = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])
    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    a1sb_a1x = np.array([da1sb_ax, da1sb_rx])
    a1sb_a2x = np.array([da1sb_2ax, da1sb_arx, da1sb_2rx])

    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x


def da1sB_dx_dxhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xs_m, zs_m,
                               I_lambdaskl, J_lambdaskl, ccteskl, a1vdwkl,
                               a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                               dxhix_dx_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lakl,
                                   J_lakl, cctes_lakl, a1vdw_lakl, a1vdw_ctekl,
                                   dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1sb_a, da1sb_a, da1sb_ax, da1sb_axxhi = out_la
    out_lr = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lrkl,
                                   J_lrkl, cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl,
                                   dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1sb_r, da1sb_r, da1sb_rx, da1sb_rxxhi = out_lr
    out_2la = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lakl,
                                    J_2lakl, cctes_2lakl, a1vdw_2lakl,
                                    a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_2a, da1sb_2a, da1sb_2ax, da1sb_2axxhi = out_2la
    out_2lr = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_2lrkl,
                                    J_2lrkl, cctes_2lrkl, a1vdw_2lrkl,
                                    a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_2r, da1sb_2r, da1sb_2rx, da1sb_2rxxhi = out_2lr
    out_lar = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_larkl,
                                    J_larkl, cctes_larkl, a1vdw_larkl,
                                    a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_ar, da1sb_ar, da1sb_arx, da1sb_arxxhi = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    a1sb_a1x = np.array([da1sb_ax, da1sb_rx])
    a1sb_a2x = np.array([da1sb_2ax, da1sb_arx, da1sb_2rx])

    a1sb_a1xxhi = np.array([da1sb_axxhi, da1sb_rxxhi])
    a1sb_a2xxhi = np.array([da1sb_2axxhi, da1sb_arxxhi, da1sb_2rxxhi])

    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x, a1sb_a1xxhi, a1sb_a2xxhi


def da1sB_dx_d2xhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                I_lambdaskl, J_lambdaskl, ccteskl, a1vdwkl,
                                a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                dxhix_dx_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lakl,
                                    J_lakl, cctes_lakl, a1vdw_lakl,
                                    a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a, da1sb_ax, da1sb_axxhi = out_la
    out_lr = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m, I_lrkl,
                                    J_lrkl, cctes_lrkl, a1vdw_lrkl,
                                    a1vdw_ctekl, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r, da1sb_rx, da1sb_rxxhi = out_lr
    out_2la = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                     I_2lakl, J_2lakl, cctes_2lakl,
                                     a1vdw_2lakl, a1vdw_ctekl, dxhix_dxhi00,
                                     dxhix_dx, dxhix_dx_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a, da1sb_2ax, da1sb_2axxhi = out_2la
    out_2lr = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                     I_2lrkl, J_2lrkl, cctes_2lrkl,
                                     a1vdw_2lrkl, a1vdw_ctekl, dxhix_dxhi00,
                                     dxhix_dx, dxhix_dx_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r, da1sb_2rx, da1sb_2rxxhi = out_2lr
    out_lar = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                     I_larkl, J_larkl, cctes_larkl,
                                     a1vdw_larkl, a1vdw_ctekl, dxhix_dxhi00,
                                     dxhix_dx, dxhix_dx_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar, da1sb_arx, da1sb_arxxhi = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                        [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])

    a1sb_a1x = np.array([da1sb_ax, da1sb_rx])
    a1sb_a2x = np.array([da1sb_2ax, da1sb_arx, da1sb_2rx])

    a1sb_a1xxhi = np.array([da1sb_axxhi, da1sb_rxxhi])
    a1sb_a2xxhi = np.array([da1sb_2axxhi, da1sb_arxxhi, da1sb_2rxxhi])

    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x, a1sb_a1xxhi, a1sb_a2xxhi


def x0lambda_evalm(x0, laij, lrij, larij):
    x0la = x0**laij
    x0lr = x0**lrij
    x02la = x0**(2*laij)
    x02lr = x0**(2*lrij)
    x0lar = x0**larij

    # To be used for a1 and a2 in monomer term
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])

    return x0_a1, x0_a2


def x0lambda_evalc(x0, la, lr, lar):
    x0la = x0**la
    x0lr = x0**lr
    x02la = x0**(2*la)
    x02lr = x0**(2*lr)
    x0lar = x0**lar

    # To be used for a1 and a2 in monomer term
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])
    # To be used in g1 and g2 of chain
    x0_g1 = np.array([la * x0la, -lr*x0lr])
    x0_g2 = np.array([la * x02la, -lar*x0lar, lr * x02lr])

    return x0_a1, x0_a2, x0_g1, x0_g2



from __future__ import division, print_function, absolute_import

# Equation 64
def g1sigma(rho, suma_g1, da1m_deta, deta_drho, cte_g1s):

    da1 = da1m_deta * deta_drho
    g1 = cte_g1s * (3.*da1 - suma_g1/rho)

    return g1


def dg1sigma_drho(rho, suma_g1, d2a1m_drho, cte_g1s):
    dg1 = 3.*d2a1m_drho - suma_g1/rho
    dg1[1] += suma_g1[0]/rho**2
    dg1 *= cte_g1s

    return dg1


def d2g1sigma_drho(rho, suma_g1, d3a1m_drho, cte_g1s):

    rho2 = rho**2
    rho3 = rho2*rho
    d2g1 = 3*d3a1m_drho - suma_g1/rho
    d2g1[1] += suma_g1[0]/rho2
    d2g1[2] += -2*suma_g1[0]/rho3 + 2*suma_g1[1]/rho2
    d2g1 *= cte_g1s

    return d2g1



from __future__ import division, print_function, absolute_import
import numpy as np
import pandas as pd
import os
from copy import copy

databasepath = os.path.join(os.path.dirname(__file__), 'database')
databasepath += '/saftgamma_database.xlsx'

file = pd.ExcelFile(databasepath, engine='openpyxl')
df_groups = pd.read_excel(file, 'groups', index_col='groups')
df_mie_kl = pd.read_excel(file, 'unlikemie_kl')
df_asso_kl = pd.read_excel(file, 'unlikeasso_kl')
df_secondorder = pd.read_excel(file, 'secondmie')
df_secondasso = pd.read_excel(file, 'secondasso')


class GCdatabase(object):
    '''
    SAFT-Gamma-Mie group contribution database Object

    This object have implemeted methods to modify the database published
    groups and interactions parameters for SAFT-Gamma-Mie EoS.

    The included parameters are obtained from J. Chem. Eng. Data 2020,
    65, 12, 5862–5890. https://doi.org/10.1021/acs.jced.0c00746

    Parameters
    ----------
    df_groups: DataFrame
        DataFrame that includes group information
    df_mie_kl: DataFrame
        DataFrame that includes unlike Mie parameters
    df_asso_kl: DataFrame
        DataFrame that includes unlike association Parameters
    df_secondorder: DataFrame
        Dataframe that includes second order modification to Mie interactions
    df_scondasso: DataFrame
        Dataframe that includes second order modification to association
        interactions

    Attributes
    ----------
    group_list: List of groups on the database
    df_groups: DataFrame that includes group information
    df_mie_kl: DataFrame that includes unlike Mie parameters
    df_asso_kl: DataFrame that includes unlike association Parameters
    df_secondorder: Dataframe that includes second order modification
    to Mie interactions
    df_scondasso: Dataframe that includes second order modification to
    association interactions

    Methods
    -------
    add_group: method to add a new group
    new_interaction_mie: method to set unlike Mie interactions
    new_interaction_asso: method to add association parameters
    get_interactions: method to get the interactions between two groups
    restore_database: method to restore the database to its initial value
    '''
    def __init__(self, df_groups=df_groups, df_mie_kl=df_mie_kl,
                 df_asso_kl=df_asso_kl, df_secondorder=df_secondorder,
                 df_secondasso=df_secondasso):

        self.df_groups = df_groups
        self.df_mie_kl = df_mie_kl
        self.df_asso_kl = df_asso_kl
        self.df_secondorder = df_secondorder
        self.df_secondasso = df_secondasso

        self.df_groups_backup = copy(df_groups)
        self.df_mie_kl_backup = copy(df_mie_kl)
        self.df_asso_kl_backup = copy(df_asso_kl)
        self.df_secondorder_backup = copy(df_secondorder)
        self.df_secondasso_backup = copy(df_secondasso)

        self.group_list = list(self.df_groups.index)

    def add_group(self, name, vk=1., Sk=1., sigma=0., eps=0., lr=12., la=6.,
                  nH=0, ne1=0, ne2=0, charge=0., sigma_born=0., mw=0.,
                  author_key='author', doi='doi',
                  overwrite=False):
        """
        add_group method

        Method that adds a new group to the database

        Parameters
        ----------
        name: string
            name of the group
        vk: float
            number os sites used by the group
        Sk: float
            shape factor of the group
        sigma: float
            lenght scale of the group, used in Mie potential [Amstrong]
        eps: float
            energy scale of the group, used in Mie potential [K]
        lr: float
            repulsive exponent of the group, used in Mie potential
        la: float
            attractive exponent of the group, used in Mie potential
        nH: int
            number of Hidrogen associative sites
        ne1: int
            number of e1 associative sites
        ne2: int
            number of e2 associative sites
        charge: int
            charge of the group (electron charge) [Adim.]
        sigma_born: float
            diameter used in Born contribution [Amstrong]
        mw: float
            molar weight of the group [g/mol]
        author_key: string
            key of the author that provided the parameters
        doi: string
            doi of the publication that provided the parameters
        overwrite: bool, optional
            whether to overwrite or not current parameters of the database
        """
        Nst = np.count_nonzero([nH, ne1, ne2])
        group_included = name in self.group_list

        if group_included and not overwrite:
            raise Exception('group {} already included in database, set'.format(name) +
                            ' overwrite=True to overwrite the current parameters')
        elif group_included and overwrite:
            new_parameters = [vk, Sk, sigma, eps, lr, la, Nst, nH,
                              ne1, ne2, charge, sigma_born, mw,
                              author_key, doi]

            self.df_groups.loc[name] = new_parameters
        else:
            new_group = {'groups': name, 'vk*': vk, 'Sk': Sk,
                         'sigma_kk': sigma, 'eps_kk': eps,
                         'lr_kk': lr, 'la_kk': la, 'Nst_kk': Nst, 'nH_kk': nH,
                         'ne1_kk': ne1, 'ne2_kk': ne2, 'charge_kk': charge,
                         'sigma_born_kk': sigma_born, 'mw_kk': mw,
                         'author_key': author_key, 'doi': doi}
            groups_aux = self.df_groups.reset_index()
            groups_new = groups_aux.append(new_group, ignore_index=True)
            groups_new.set_index('groups', inplace=True)
            self.df_groups = groups_new
            self.group_list = list(self.df_groups.index)

    def new_interaction_mie(self, group_k, group_l, eps_kl=0., lr_kl='CR',
                            author_key='author', doi='doi',
                            overwrite=False):
        """
        new_interaction_mie method

        Method that adds the unlike mie interactions between group k and
        group l

        Parameters
        ----------
        group_k: string
            name of the group k
        group_l: string
            name of the group l
        eps_kl: float
            unlike energy scale between the groups, used in Mie potential [K]
        lr_kl: float
            unlike repulsive exponent between the groups, used in Mie potential
        author_key: string
            key of the author that provided the parameters
        doi: string
            doi of the publication that provided the parameters
        overwrite: bool, optional
            whether to overwrite or not current parameters of the database
        """
        bool_kk = self.df_mie_kl.group_k == group_k
        bool_ll = self.df_mie_kl.group_l == group_l
        bool_kl = self.df_mie_kl.group_k == group_l
        bool_lk = self.df_mie_kl.group_l == group_k

        index1 = self.df_mie_kl.index[bool_kk & bool_ll]
        len1 = index1.shape[0]

        index2 = self.df_mie_kl.index[bool_kl & bool_lk]
        len2 = index2.shape[0]

        if len1 == 1:
            index = index1
            n = len1
        elif len2 == 1:
            index = index2
            n = len2
        else:
            n = 0
        already_in = n > 0
        if already_in and not overwrite:
            raise Exception('Interaction parameters between group {} '.format(group_k) +
                            ' and group {}'.format(group_l) + ' are already in the database' +
                            ' set overwrite=True to overwrite the current parameters')

        elif already_in and overwrite:
            self.df_mie_kl.iloc[index, [2, 3, 4, 5]] = [eps_kl, lr_kl, author_key, doi]
        else:
            new_interaction = {'group_k': group_k, 'group_l': group_l,
                               'eps_kl': eps_kl, 'lr_kl': lr_kl,
                               'author_key': author_key, 'doi': doi}
            df_mie_new = self.df_mie_kl.append(new_interaction, ignore_index=True)
            self.df_mie_kl = df_mie_new

    def new_interaction_asso(self, group_k, group_l, site_k, site_l,
                             epsAB_kl=0., kAB_kl=0,
                             author_key='author', doi='doi',
                             overwrite=False):
        """
        new_interaction_mie method

        Method that adds the unlike mie interactions between group k and
        group l

        Parameters
        ----------
        group_k: string
            name of the group k
        group_l: string
            name of the group l
        site_k: string
            name of the site k, available optiones are 'H', 'e1' and 'e2'
        site_l: string
            name of the site l, available optiones are 'H', 'e1' and 'e2'
        epsAB_kl: float
            unlike association energy between the the groups [K]
        kAB_kl: float
            unlike association volume between the groups [Amstrong^3]
        author_key: string
            key of the author that provided the parameters
        doi: string
            doi of the publication that provided the parameters
        overwrite: bool, optional
            whether to overwrite or not current parameters of the database
        """
        bool_kk = self.df_asso_kl.group_k == group_k
        bool_ll = self.df_asso_kl.group_l == group_l
        boolsite_kk = self.df_asso_kl.iloc[:, 1] == site_k
        boolsite_ll = self.df_asso_kl.iloc[:, 3] == site_l

        bool_kl = self.df_asso_kl.group_k == group_l
        bool_lk = self.df_asso_kl.group_l == group_k
        boolsite_kl = self.df_asso_kl.iloc[:, 1] == site_l
        boolsite_lk = self.df_asso_kl.iloc[:, 3] == site_k

        index1 = self.df_asso_kl.index[bool_kk & bool_ll & boolsite_kk & boolsite_ll]
        len1 = index1.shape[0]

        index2 = self.df_asso_kl.index[bool_kl & bool_lk & boolsite_kl & boolsite_lk]
        len2 = index2.shape[0]

        if len1 == 1:
            index = index1
            n = len1
        elif len2 == 1:
            index = index2
            n = len2
        else:
            n = 0

        already_in = n > 0
        if already_in and not overwrite:
            raise Exception('Interaction parameters between site{} of'.format(site_k) +
                            ' of group {} '.format(group_k) + 'and site {}'.format(site_l) +
                            ' of group {}'.format(group_l) + 'are already in the database' +
                            ' set overwrite=True to overwrite the current parameters')

        elif already_in and overwrite:
            self.df_asso_kl.iloc[index, [4, 5, 6, 7]] = [epsAB_kl, kAB_kl, author_key, doi]
        else:
            new_interaction = {'group_k': group_k,
                               'site\xa0a\xa0of group\xa0k': site_k,
                               'group_l': group_l,
                               'site\xa0b\xa0of group\xa0l': site_l,
                               'epsAB_kl': epsAB_kl, 'KAB_kl': kAB_kl,
                               'author_key': author_key, 'doi': doi}
            df_asso_new = self.df_asso_kl.append(new_interaction, ignore_index=True)
            self.df_asso_kl = df_asso_new

    def get_interactions(self, group_k, group_l):
        """
        get_interactions method

        Method that outputs the available interactions between group k and
        group l

        Parameters
        ----------
        group_k: string
            name of the group k
        group_l: string
            name of the group l

        Returns
        -------
        df_group : DataFrame
             parameters of each group
        df_mie : DataFrame
             unlike Mie Parameters
        df_asso : DataFrame
             unlike association parameters
        """
        df_group = self.df_groups.loc[[group_l, group_k]]

        bool_kk = self.df_mie_kl.group_k == group_k
        bool_ll = self.df_mie_kl.group_l == group_l
        bool_kl = self.df_mie_kl.group_k == group_l
        bool_lk = self.df_mie_kl.group_l == group_k

        df1 = self.df_mie_kl[bool_kk & bool_ll]
        len1 = df1.shape[0]

        df2 = self.df_mie_kl[bool_kl & bool_lk]
        len2 = df2.shape[0]

        if len1 == 1:
            df_mie = df1
        elif len2 == 1:
            df_mie = df2
        else:
            df_mie = 'There are no custom interaction parameters set for group {}'.format(group_k)
            df_mie += 'and group {}'.format(group_l)

        bool_kk = self.df_asso_kl.group_k == group_k
        bool_ll = self.df_asso_kl.group_l == group_l
        bool_kl = self.df_asso_kl.group_k == group_l
        bool_lk = self.df_asso_kl.group_l == group_k

        df1 = self.df_asso_kl[bool_kk & bool_ll]
        len1 = df1.shape[0]

        df2 = self.df_asso_kl[bool_kl & bool_lk]
        len2 = df2.shape[0]

        if len1 >= 1:
            df_asso = df1
        elif len2 >= 1:
            df_asso = df2
        else:
            df_asso = 'There are no association parameters set for group {}'.format(group_k)
            df_asso += ' and group {}'.format(group_l)

        return df_group, df_mie, df_asso

    def restore_database(self):
        """
        restore_database method

        Method that restores the database of its initial state.
        This method will erase any custom groups or interactions
        set up by the user.
        """
        self.df_groups = copy(self.df_groups_backup)
        self.df_mie_kl = copy(self.df_mie_kl_backup)
        self.df_asso_kl = copy(self.df_asso_kl_backup)
        self.df_secondorder = copy(self.df_secondorder_backup)
        self.df_secondasso = copy(self.df_secondasso_backup)

        self.group_list = list(self.df_groups.index)


database = GCdatabase()



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from ..math import gdem
from .equilibriumresult import EquilibriumResult


def bubble_sus(P_T, X, T_P, bubble_type, y_guess, eos,
               vl0, vv0, Xassl0, Xassv0,
               not_in_y=[]):

    if bubble_type == 'T':
        P = P_T
        temp_aux = T_P
    elif bubble_type == 'P':
        temp_aux = P_T
        P = T_P

    # Liquid fugacities
    lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L', vl0, Xassl0)
    if np.any(np.isnan(lnphil)) or np.isnan(vl):
        lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L')

    tol = 1e-8
    error = 1
    itacc = 0
    niter = 0
    n = 5
    Y_calc = y_guess
    Y = y_guess

    # Vapour fugacities
    lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V', vv0, Xassv0)
    if np.any(np.isnan(lnphiv)) or np.isnan(vv):
        lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V')

    while error > tol and itacc < 3:
        niter += 1

        lnK = lnphil-lnphiv
        K = np.exp(lnK)
        Y_calc_old = Y_calc
        Y_calc = X*K

        # modification for non-volatile components Ki -> 0
        Y_calc[not_in_y] = 0.

        if niter == (n-3):
            Y3 = Y_calc
        elif niter == (n-2):
            Y2 = Y_calc
        elif niter == (n-1):
            Y1 = Y_calc
        elif niter == n:
            niter = 0
            itacc += 1
            dacc = gdem(Y_calc, Y1, Y2, Y3)
            Y_calc += dacc
        error = np.linalg.norm(Y_calc-Y_calc_old)
        Y = Y_calc/Y_calc.sum()
        # Vapor fugacities
        lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V', vv, Xassv)
        if np.any(np.isnan(lnphiv)) or np.isnan(vv):
            lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V')

    if bubble_type == 'T':
        f0 = Y_calc.sum() - 1
    elif bubble_type == 'P':
        f0 = np.log(Y_calc.sum())

    return f0, Y, lnK, vl, vv, Xassl, Xassv


def bubble_newton(inc, X, T_P, bubble_type, eos, in_y=[]):
    global vl, vv, Xassl, Xassv
    f = np.zeros_like(inc)
    lnK = inc[:-1]
    K = np.exp(lnK)

    if bubble_type == 'T':
        P = inc[-1]
        temp_aux = T_P
    elif bubble_type == 'P':
        T = inc[-1]
        temp_aux = eos.temperature_aux(T)
        P = T_P

    nc = eos.nc
    Y = np.zeros(nc)
    Y[in_y] = X[in_y] * K

    # Liquid fugacities
    lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L', vl, Xassl)
    if np.any(np.isnan(lnphil)) or np.isnan(vl):
        lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L')
    # Vapour fugacities
    lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V', vv, Xassv)
    if np.any(np.isnan(lnphiv)) or np.isnan(vv):
        lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V')

    f[:-1] = lnK + (lnphiv - lnphil)[in_y]
    f[-1] = (Y-X).sum()

    return f


def bubblePy(y_guess, P_guess, X, T, model, good_initial=False,
             v0=[None, None], Xass0=[None, None],
             not_in_y_list=[],
             full_output=False):
    """
    Bubble point (T, x) -> (P, y)

    Solves bubble point at given liquid composition and temperature. It uses a
    combination of accelerated successive sustitution with quasi Newton Method
    in regular cases and when good initial it's provided the full system of
    equations of the phase envelope method is used as objective function.

    Parameters
    ----------
    y_guess : array_like
        guess of vapour phase composition
    P_guess : float
        guess of equilibrium pressure [Pa].
    X : array_like
        liquid phase composition
    T : float
        absolute temperature of the liquid [K].
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip succesive substitution and solves by Newton's Method.
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    Xass0 : list, optional
        if supplied X association non-bonded fraction sites initial value
    not_in_y_list : list, optional
        index of components not present in the vapour phase (default is empty)
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    Y : array_like, vector of vapour fraction moles
    P : float, equilibrium pressure [Pa]

    """
    nc = model.nc
    if len(y_guess) != nc or len(X) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    if np.any([i > (nc-1) for i in not_in_y_list]):
        raise Exception('Index of components not_in_y_list must be less than nc')

    not_in_y = np.zeros(nc, dtype=bool)
    not_in_y[not_in_y_list] = 1
    in_y = np.logical_not(not_in_y)

    global vl, vv, Xassl, Xassv
    vl0, vv0 = v0
    Xassl0, Xassv0 = Xass0

    temp_aux = model.temperature_aux(T)

    it = 0
    itmax = 10
    tol = 1e-8

    P = P_guess
    out = bubble_sus(P, X, temp_aux, 'T', y_guess, model, vl0, vv0,
                     Xassl0, Xassv0, not_in_y=not_in_y)
    f, Y, lnK, vl, vv, Xassl, Xassv = out
    error = np.abs(f)
    h = 1e-4

    method = 'quasi-newton + ASS'
    while error > tol and it <= itmax and not good_initial:
        it += 1
        out = bubble_sus(P+h, X, temp_aux, 'T', Y, model, vl, vv,
                         Xassl, Xassv, not_in_y=not_in_y)
        f1, Y1, lnK1, vl, vv, Xassl, Xassv = out
        out = bubble_sus(P, X, temp_aux, 'T', Y, model, vl, vv,
                         Xassl, Xassv, not_in_y=not_in_y)
        f, Y, lnK, vl, vv, Xassl, Xassv = out
        df = (f1-f)/h
        dP = f / df
        if dP > P:
            dP = 0.4 * P
        elif np.isnan(dP):
            dP = 0.0
            it = 1.*itmax
        P -= dP
        error = np.abs(f)

    if error > tol:
        inc0 = np.hstack([lnK[in_y], P])
        sol1 = root(bubble_newton, inc0, args=(X, temp_aux, 'T', model, in_y))
        if sol1.success:
            method = 'second-order'
            error = np.linalg.norm(sol1.fun)
            it += sol1.nfev

            sol = sol1.x
            lnK_newton = sol[:-1]
            K_newton = np.exp(lnK_newton)
            Y = np.zeros(nc)
            Y[in_y] = X[in_y] * K_newton
            Y /= np.sum(Y)
            P = sol[-1]

            rhol, Xassl = model.density_aux(X, temp_aux, P, 'L', rho0=1./vl,
                                            Xass0=Xassl)
            rhov, Xassv = model.density_aux(Y, temp_aux, P, 'V', rho0=1./vv,
                                            Xass0=Xassv)
            vl = 1./rhol
            vv = 1./rhov

    if full_output:
        sol = {'T': T, 'P': P, 'error': error, 'iter': it,
               'X': X, 'v1': vl, 'Xassl': Xassl, 'state1': 'Liquid',
               'Y': Y, 'v2': vv, 'Xassv': Xassv, 'state2': 'Vapor',
               'method': method}
        out = EquilibriumResult(sol)
        return out

    return Y, P


def bubbleTy(y_guess, T_guess, X, P, model, good_initial=False,
             v0=[None, None], Xass0=[None, None],
             not_in_y_list=[],
             full_output=False):
    """
    Bubble point (P, x) -> (T, y)

    Solves bubble point at given liquid composition and pressure. It uses a
    combination of accelerated successive sustitution with quasi Newton Method
    in regular cases and when good initial it's provided the full system of
    equations of the phase envelope method is used as objective function.

    Parameters
    ----------
    y_guess : array_like
        guess of vapour phase composition
    T_guess : float
        guess of equilibrium temperature of the liquid [K].
    X : array_like
        liquid phase composition
    P : float
        pressure of the liquid [Pa]
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip succesive substitution and solves by Newton's Method.
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    Xass0 : list, optional
        if supplied X association non-bonded fraction sites initial value
    not_in_y_list : list, optional
        index of components not present in the vapour phase (default is empty)
    full_output: bool, optional
        wheter to outputs all calculation info


    Returns
    -------
    Y : array_like
        vector of vapour fraction moles
    T : float
        equilibrium temperature [K]
    """

    nc = model.nc
    if len(y_guess) != nc or len(X) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    if np.any([i > (nc-1) for i in not_in_y_list]):
        raise Exception('Index of components not_in_y_list must be less than nc')

    not_in_y = np.zeros(nc, dtype=bool)
    not_in_y[not_in_y_list] = 1
    in_y = np.logical_not(not_in_y)

    global vl, vv, Xassl, Xassv
    vl0, vv0 = v0
    Xassl0, Xassv0 = Xass0

    it = 0
    itmax = 10
    tol = 1e-8

    T = T_guess
    temp_aux = model.temperature_aux(T)
    out = bubble_sus(temp_aux, X, P, 'P', y_guess, model, vl0, vv0, Xassl0,
                     Xassv0, not_in_y=not_in_y)
    f, Y, lnK, vl, vv, Xassl, Xassv = out
    # f, Y, lnK, vl, vv = bubble_sus(T, X, P, 'P', y_guess, model, vl0, vv0)
    error = np.abs(f)
    h = 1e-4
    method = 'quasi-newton + ASS'
    while error > tol and it <= itmax and not good_initial:
        it += 1
        temp_aux = model.temperature_aux(T+h)
        out = bubble_sus(temp_aux, X, P, 'P', Y, model, vl, vv, 
                         Xassl, Xassv, not_in_y=not_in_y)
        f1, Y1, lnK1, vl, vv, Xassl, Xassv = out
        temp_aux = model.temperature_aux(T)
        out = bubble_sus(temp_aux, X, P, 'P', Y, model, vl, vv, 
                         Xassl, Xassv, not_in_y=not_in_y)
        f, Y, lnK, vl, vv, Xassl, Xassv = out
        df = (f1-f)/(h)
        T -= f/df
        error = np.abs(f)

    if error > tol:
        inc0 = np.hstack([lnK[in_y], T])
        sol1 = root(bubble_newton, inc0, args=(X, P, 'P', model, in_y))
        if sol1.success:
            method = 'second-order'
            error = np.linalg.norm(sol1.fun)
            it += sol1.nfev

            sol = sol1.x
            lnK_newton = sol[:-1]
            K_newton = np.exp(lnK_newton)
            Y = np.zeros(nc)
            Y[in_y] = X[in_y] * K_newton
            Y /= np.sum(Y)
            T = sol[-1]

            rhol, Xassl = model.density_aux(X, temp_aux, P, 'L', rho0=1./vl,
                                            Xass0=Xassl)
            rhov, Xassv = model.density_aux(Y, temp_aux, P, 'V', rho0=1./vv,
                                            Xass0=Xassv)
            vl = 1./rhol
            vv = 1./rhov

    if full_output:
        sol = {'T': T, 'P': P, 'error': error, 'iter': it,
               'X': X, 'v1': vl, 'Xassl': Xassl, 'state1': 'Liquid',
               'Y': Y, 'v2': vv, 'Xassv': Xassv, 'state2': 'Vapor',
               'method': method}
        out = EquilibriumResult(sol)
        return out

    return Y, T


__all__ = ['bubbleTy', 'bubblePy']


from __future__ import division, print_function, absolute_import
import numpy as np


def g1sigma(xhi00, xm, da1_chain, suma_g1, a1vdw_cteii):
    g1 = 3 * da1_chain - suma_g1/xhi00
    g1 /= xm
    g1 /= -a1vdw_cteii
    return g1


def dg1sigma_dxhi00(xhi00, xm, d2a1_chain, dsuma_g1, a1vdw_cteii):

    sum1, dsum1 = dsuma_g1
    g1 = 3 * d2a1_chain
    g1[0] -= sum1/xhi00
    g1[1] += (sum1/xhi00 - dsum1)/xhi00
    g1 /= xm
    g1 /= -a1vdw_cteii

    return g1


def d2g1sigma_dxhi00(xhi00, xm, d3a1_chain, d2suma_g1, a1vdw_cteii):

    sum1, dsum1, d2sum1 = d2suma_g1
    g1 = 3. * d3a1_chain - d2suma_g1/xhi00
    g1[1] += sum1/xhi00**2
    g1[2] += -2*sum1/xhi00**3 + 2*dsum1/xhi00**2
    g1 /= xm
    g1 /= -a1vdw_cteii
    return g1


def dg1sigma_dx(xhi00, xm, ms, da1, da1x, suma_g1, suma_g1x, a1vdw_cte):

    g1 = 3. * da1 - suma_g1/xhi00
    g1 /= xm
    g1 /= -a1vdw_cte

    g1x = np.multiply.outer(ms, (suma_g1/xhi00 - 3. * da1)/xm)
    g1x += -suma_g1x/xhi00
    g1x += 3*da1x
    g1x /= xm
    g1x /= -a1vdw_cte

    return g1, g1x


def dg1sigma_dxxhi(xhi00, xm, ms, da1, da1x, suma_g1, suma_g1x, a1vdw_cte):

    sum1, dsum1 = suma_g1

    g1 = 3 * da1
    g1[0] -= sum1/xhi00
    g1[1] += (sum1/xhi00 - dsum1)/xhi00
    g1 /= xm
    g1 /= -a1vdw_cte

    g1x = np.multiply.outer(ms, (sum1/xhi00 - 3. * da1[0])/xm)
    g1x += -suma_g1x/xhi00
    g1x += 3*da1x
    g1x /= xm
    g1x /= -a1vdw_cte
    return g1, g1x
